package de.fhhannover.inform.trust.ifmapj;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.fail;

import org.w3c.dom.Document;
import org.w3c.dom.NodeList;

import de.fhhannover.inform.trust.ifmapj.binding.IfmapStrings;
import de.fhhannover.inform.trust.ifmapj.channel.ARC;
import de.fhhannover.inform.trust.ifmapj.channel.SSRC;
import de.fhhannover.inform.trust.ifmapj.exception.EndSessionException;
import de.fhhannover.inform.trust.ifmapj.exception.IfmapErrorResult;
import de.fhhannover.inform.trust.ifmapj.exception.IfmapException;
import de.fhhannover.inform.trust.ifmapj.identifier.IdentifierFactory;
import de.fhhannover.inform.trust.ifmapj.identifier.IpAddress;
import de.fhhannover.inform.trust.ifmapj.identifier.MacAddress;
import de.fhhannover.inform.trust.ifmapj.messages.PollResult;
import de.fhhannover.inform.trust.ifmapj.messages.PublishDelete;
import de.fhhannover.inform.trust.ifmapj.messages.PublishRequest;
import de.fhhannover.inform.trust.ifmapj.messages.PublishUpdate;
import de.fhhannover.inform.trust.ifmapj.messages.RequestFactory;
import de.fhhannover.inform.trust.ifmapj.messages.ResultItem;
import de.fhhannover.inform.trust.ifmapj.messages.SearchResult;
import de.fhhannover.inform.trust.ifmapj.messages.SubscribeRequest;
import de.fhhannover.inform.trust.ifmapj.messages.SubscribeUpdate;
import de.fhhannover.inform.trust.ifmapj.metadata.StandardIfmapMetadataFactory;

/**
 * All the assert calls are pretty irond specific. Other MAPSs might give
 * failures...
 * 
 * @author aw
 *
 */
public class SubscriptionTest {

	private static final RequestFactory sRf;
	private static final IdentifierFactory sIf;
	private static final StandardIfmapMetadataFactory sMf;
	private static final IpAddress sIp;
	private static final MacAddress sMac; 
	static {
		sMf = IfmapJ.createStandardMetadataFactory();
		sRf = IfmapJ.createRequestFactory();
		sIf = IfmapJ.createIdentifierFactory();
		sIp = sIf.createIp4("192.168.0.1");
		sMac = sIf.createMac("aa:bb:cc:dd:ee:ff");
	}

	public static void main(String args[]) throws  IfmapErrorResult, IfmapException, EndSessionException, InterruptedException {
		
		IfmapJHelper.setSystemProperties(
				ConfData.TRUST_STORE_PATH,
				ConfData.TRUST_STORE_PASSWORD,
				ConfData.KEY_STORE_PATH,
				ConfData.KEY_STORE_PASSWORD);
		
		PollResult pr = null;
		
		SSRC ssrc = IfmapJ.createSSRC(
				ConfData.BASIC_AUTH_SERVER_URL,
				ConfData.BASIC_AUTH_USER,
				ConfData.BASIC_AUTH_PASSWORD,
				IfmapJHelper.getTrustManagers());
		final ARC arc = ssrc.getArc();
		
		ssrc.newSession();
		prepareSubscriptions(ssrc);
		
		pr = arc.poll();

		assertEquals(2, pr.getSearchResults().size());
		assertEquals(0, pr.getUpdateResults().size());
		assertEquals(0, pr.getDeleteResults().size());
		assertEquals(0, pr.getNotifyResults().size());
		for (SearchResult sr : pr.getUpdateResults()) {
			assertEquals(1, sr.getResultItems().size());
			ResultItem ri = sr.getResultItems().iterator().next();
			assertNotNull(ri.getIdentifier1());
			assertNull(ri.getIdentifier2());
			assertEquals(0, ri.getMetadata().size());
		}
		
		publishLink(ssrc);
		
		pr = arc.poll();
		
		assertEquals(0, pr.getSearchResults().size());
		assertEquals(2, pr.getUpdateResults().size());
		assertEquals(0, pr.getDeleteResults().size());
		assertEquals(0, pr.getNotifyResults().size());
		for (SearchResult sr : pr.getUpdateResults()) {
			assertEquals(1, sr.getResultItems().size());
			ResultItem ri = sr.getResultItems().iterator().next();
			assertNotNull(ri.getIdentifier1());
			assertNotNull(ri.getIdentifier2());
			assertEquals(1, ri.getMetadata().size());
		}
		publishStuff(ssrc);

		pr = arc.poll();
		
		assertEquals(0, pr.getSearchResults().size());
		assertEquals(2, pr.getUpdateResults().size());
		assertEquals(0, pr.getDeleteResults().size());
		assertEquals(0, pr.getNotifyResults().size());
		for (SearchResult sr : pr.getUpdateResults()) {
			assertEquals(2, sr.getResultItems().size());
			for (ResultItem ri : sr.getResultItems()) {
				assertNotNull(ri.getIdentifier1());
				assertNull(ri.getIdentifier2());
				assertEquals(1, ri.getMetadata().size());
				if (ri.getIdentifier1() instanceof IpAddress) {
					IpAddress ip = (IpAddress)ri.getIdentifier1();
					assertEquals(sIp.getValue(), ip.getValue());
				} else if (ri.getIdentifier1() instanceof MacAddress) {
					MacAddress mac = (MacAddress)ri.getIdentifier1();
					assertEquals(sMac.getValue(), mac.getValue());
				} else {
					fail();
				}
			}
		}
		deleteLink(ssrc);
		
		pr = arc.poll();
		assertEquals(0, pr.getSearchResults().size());
		assertEquals(0, pr.getUpdateResults().size());
		assertEquals(2, pr.getDeleteResults().size());
		assertEquals(0, pr.getNotifyResults().size());
		
		for (SearchResult sr : pr.getDeleteResults()) {
			assertEquals(2, sr.getResultItems().size());
			for (ResultItem ri : sr.getResultItems()) {
				assertEquals(1, ri.getMetadata().size());
				if (ri.holdsLink()) {
					assertNotNull(ri.getIdentifier1());
					assertNotNull(ri.getIdentifier2());
					Document doc = ri.getMetadata().iterator().next();
					NodeList list = doc.getElementsByTagNameNS(
							IfmapStrings.STD_METADATA_NS_URI, "ip-mac");
					assertEquals(1, list.getLength());
				} else {
					assertNotNull(ri.getIdentifier1());
					assertNull(ri.getIdentifier2());
					Document doc = ri.getMetadata().iterator().next();
					NodeList list = doc.getElementsByTagNameNS(
							IfmapStrings.STD_METADATA_NS_URI, "device-attribute");
					assertEquals(1, list.getLength());
					NodeList list2 = doc.getElementsByTagName("name");
					assertEquals(1, list2.getLength());
					
					if (sr.getName().equals("sub-ip")) {
						// If this is the deleteResult for the sub-ip subscription,
						// we see the macdummy metadata of the mac identifier as deleted 
					   assertEquals(list2.item(0).getTextContent(), "macdummy");
					} else if (sr.getName().equals("sub-mac")) {
						// But if this is the deleteResult for the sub-mac subscription,
						// we see the ipdummy metadata of the ip identifier as deleted 
						assertEquals(list2.item(0).getTextContent(), "ipdummy");
					} else {
						fail("Unknown subscription name");
					}
				}
			}
		}
		
		new Thread(new Runnable() {

			@Override
			public void run() {
				try {
					arc.poll();
				} catch (EndSessionException e) {
					// that's what should happen
				} catch (Exception e) {
					fail("That shouldn't happend");
				}
			}
			
		}).start();
		
		// let the arc.poll() thread run
		Thread.sleep(100);
		ssrc.endSession();
	}
	
	private static void publishLink(SSRC ssrc) throws IfmapErrorResult, IfmapException {
		PublishRequest pr = sRf.createPublishReq();
		PublishUpdate pu = sRf.createPublishUpdate();
		pu.setIdentifier1(sIp);
		pu.setIdentifier2(sMac);
		pu.addMetadata(IfmapJ.createStandardMetadataFactory().createIpMac(null,null,null));
		pr.addPublishElement(pu);
		ssrc.publish(pr);
	}

	private static void prepareSubscriptions(SSRC ssrc) throws IfmapErrorResult, IfmapException {
	SubscribeRequest sr = sRf.createSubscribeReq();
		SubscribeUpdate su1 = sRf.createSubscribeUpdate();
		SubscribeUpdate su2 = sRf.createSubscribeUpdate();
		su1.setName("sub-ip");
		su2.setName("sub-mac");
		su1.setStartIdentifier(sIp);
		su2.setStartIdentifier(sMac);
		su1.setMaxDepth(2);
		su2.setMaxDepth(2);
		
		/* omapd doesn't like a non-existing match-links filter */
		su1.setMatchLinksFilter("meta:ip-mac");
		su2.setMatchLinksFilter("meta:ip-mac");
		su1.addNamespaceDeclaration("meta", IfmapStrings.STD_METADATA_NS_URI);
		su2.addNamespaceDeclaration("meta", IfmapStrings.STD_METADATA_NS_URI);
		sr.addSubscribeElement(su1);
		sr.addSubscribeElement(su2);
		ssrc.subscribe(sr);
	}
	
	public static void publishStuff(SSRC ssrc) throws IfmapErrorResult, IfmapException {
		PublishRequest pr = sRf.createPublishReq();
		PublishUpdate pu1 = sRf.createPublishUpdate();
		PublishUpdate pu2 = sRf.createPublishUpdate();
		pu1.setIdentifier1(sIp);
		pu1.addMetadata(sMf.createDevAttr("ipdummy"));
		pu2.setIdentifier2(sMac);
		pu2.addMetadata(sMf.createDevAttr("macdummy"));
		pr.addPublishElement(pu1);
		pr.addPublishElement(pu2);
		ssrc.publish(pr);
		
	}

	private static void deleteLink(SSRC ssrc) throws IfmapErrorResult, IfmapException {
		PublishRequest pr = sRf.createPublishReq();
		PublishDelete pd = sRf.createPublishDelete();
		pd.setIdentifier1(sMac);
		pd.setIdentifier2(sIp);
		pd.setFilter("meta:ip-mac");
		pd.addNamespaceDeclaration("meta", IfmapStrings.STD_METADATA_NS_URI);
		pr.addPublishElement(pd);
		ssrc.publish(pr);
	}
}
