package de.fhhannover.inform.trust.ifmapj.binding.dom;

import java.io.InputStream;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import de.fhhannover.inform.trust.ifmapj.IfmapJ;
import de.fhhannover.inform.trust.ifmapj.binding.IfmapStrings;
import de.fhhannover.inform.trust.ifmapj.binding.Unmarshaller;
import de.fhhannover.inform.trust.ifmapj.exception.EndSessionException;
import de.fhhannover.inform.trust.ifmapj.exception.IfmapErrorCode;
import de.fhhannover.inform.trust.ifmapj.exception.IfmapErrorResult;
import de.fhhannover.inform.trust.ifmapj.exception.UnmarshalException;
import de.fhhannover.inform.trust.ifmapj.identifier.AccessRequest;
import de.fhhannover.inform.trust.ifmapj.identifier.Device;
import de.fhhannover.inform.trust.ifmapj.identifier.Identifier;
import de.fhhannover.inform.trust.ifmapj.identifier.IdentifierFactory;
import de.fhhannover.inform.trust.ifmapj.identifier.IdentifierWithAd;
import de.fhhannover.inform.trust.ifmapj.identifier.Identity;
import de.fhhannover.inform.trust.ifmapj.identifier.IdentityType;
import de.fhhannover.inform.trust.ifmapj.identifier.IpAddress;
import de.fhhannover.inform.trust.ifmapj.identifier.IpAddressType;
import de.fhhannover.inform.trust.ifmapj.identifier.MacAddress;
import de.fhhannover.inform.trust.ifmapj.log.IfmapJLog;
import de.fhhannover.inform.trust.ifmapj.messages.NewSessionResult;
import de.fhhannover.inform.trust.ifmapj.messages.PollResult;
import de.fhhannover.inform.trust.ifmapj.messages.ResultFactory;
import de.fhhannover.inform.trust.ifmapj.messages.ResultFactoryImpl;
import de.fhhannover.inform.trust.ifmapj.messages.ResultItem;
import de.fhhannover.inform.trust.ifmapj.messages.SearchResult;

public class DomUnmarshaller implements Unmarshaller {

	private static DocumentBuilderFactory sDocumentBuilderFactory;
	
	private DocumentBuilder mDocumentBuilder;
	
	private IdentifierFactory mIdentifierFactory;
	
	private ResultFactory mResultFactory;
	
	static {
		sDocumentBuilderFactory = DocumentBuilderFactory.newInstance();
		sDocumentBuilderFactory.setNamespaceAware(true);
	}
	
	public DomUnmarshaller() {
		try {
			mDocumentBuilder = sDocumentBuilderFactory.newDocumentBuilder();
		} catch (ParserConfigurationException e) {
			IfmapJLog.error("Could not create DocumentBuilder instance [" 
					+ e.getMessage() + "]");
		}
		mIdentifierFactory = IfmapJ.createIdentifierFactory();
		mResultFactory = new ResultFactoryImpl();
	}

	@Override
	public NewSessionResult unmarshalNewSessionResult(InputStream is)
			throws IfmapErrorResult, UnmarshalException {
		
		Element resp = getResponseContentErrorCheck(is);
		
		return parseNewSessionResultElement(resp);
	}
	
	@Override
	public PollResult unmarshalPollResult(InputStream is)
			throws IfmapErrorResult, UnmarshalException, EndSessionException {
		Element resp = getResponseContentErrorCheck(is);
		
		return parsePollResultElement(resp);
	}
	

	@Override
	public SearchResult unmarshalSearchResult(InputStream is)
			throws IfmapErrorResult, UnmarshalException {
		Element resp = getResponseContentErrorCheck(is);
		return parseSearchResult(resp);
	}

	private SearchResult parseSearchResult(Element resp) throws UnmarshalException {
		
		SearchResult ret = null;
		
		if (!elementMatches(resp, IfmapStrings.SEARCH_RES_EL_NAME))
			throw new UnmarshalException("No " + IfmapStrings.SEARCH_RES_EL_NAME +
					" found");
		
		ret = transformSearchResult(resp);
		
		if (ret.getName() != null)
			throw new UnmarshalException("searchResult for standard search" +
					" has name \" " + ret.getName() + "\" set");
		return ret;
	}

	private SearchResult transformSearchResult(Element sres) throws UnmarshalException {
		NodeList children = sres.getChildNodes();
		Element child = null;
		List<ResultItem> ritems = new LinkedList<ResultItem>();
		String name = null;
		
		for (int i = 0; i < children.getLength(); i++) {
			
			// our interest is in elements only
			if (!(children.item(i).getNodeType() == Node.ELEMENT_NODE))
				continue;
			
			child = (Element)children.item(i);
			
			if (!elementMatches(child, IfmapStrings.RESULT_ITEM_EL_NAME))
				throw new UnmarshalException("Found non " + 
						IfmapStrings.RESULT_ITEM_EL_NAME +
						" element in SearchResult");
			
			ritems.add(transformResultItem(child));
			
		}
		/* check if a name is set */
		name = sres.getAttribute(IfmapStrings.SEARCH_RES_NAME_ATTR);
		name = (name != null && name.length() > 0) ? name : null;
		
		return mResultFactory.createSearchRes(ritems, name);
	}

	private ResultItem transformResultItem(Element el) throws UnmarshalException {
		NodeList children = el.getChildNodes();
		Identifier ident1 = null;
		Identifier ident2 = null;
		List<Document> mdlist = null;
		Element child = null;
		int childrenCount = 0;
		
		if (children.getLength() == 0)
			throw new UnmarshalException("No elements in resultItem found");
		
		for (int i = 0; i < children.getLength(); i++) {
			Identifier tmp = null;
			
			// our interest is in elements only
			if (!(children.item(i).getNodeType() == Node.ELEMENT_NODE))
				continue;
			
			child = (Element)children.item(i);
			childrenCount++;
			
			if (childrenCount > 3)
				throw new UnmarshalException("Too many elements in resultItem " +
						"element");
				
			
			if (elementMatches(child, IfmapStrings.ACCESS_REQUEST_EL_NAME)) {
				tmp = transformAr(child);
			} else if (elementMatches(child, IfmapStrings.DEVICE_EL_NAME)) {
				tmp = transformDev(child);
			} else if (elementMatches(child, IfmapStrings.IDENTITY_EL_NAME)) {
				tmp = transformIdentity(child);
			} else if (elementMatches(child, IfmapStrings.IP_ADDRESS_EL_NAME)) {
				tmp = transformIp(child);
			} else if (elementMatches(child, IfmapStrings.MAC_ADDRESS_EL_NAME)) {
				tmp = transformMac(child);
			} else if (elementMatches(child, IfmapStrings.SEARCH_METADATA_EL_NAME)) {
				if (mdlist != null)
					throw new UnmarshalException("Multiple metadata elements found");
				
				mdlist = transformMetadata(child);
			} else {
				throw new UnmarshalException("Unexpected element in searchResult");
			}
			
			if (tmp != null && ident1 == null) {
				ident1 = tmp;
			} else if (tmp != null && ident2 == null) {
				ident2 = tmp;
			} else if (tmp != null) {
				throw new UnmarshalException("Found > 2 identifiers in resultItem");
			}
		}
		
		if (ident1 == null && ident2 == null)
			throw new UnmarshalException("No identifier found in searchResult");
	
		/* in case there was no metadata element, create one */
		if (mdlist == null)
			mdlist = new LinkedList<Document>();
		
		return mResultFactory.createResultItem(ident1, ident2, mdlist);
	}


	private Identifier transformAr(Element domAr) throws UnmarshalException {
		AccessRequest ar = mIdentifierFactory.createAr();
		
		String name = domAr.getAttribute(IfmapStrings.ACCESS_REQUEST_ATTR_NAME);
		
		if (name == null)
			throw new UnmarshalException("No " + IfmapStrings.ACCESS_REQUEST_ATTR_NAME +
					" found for " + IfmapStrings.ACCESS_REQUEST_EL_NAME);
		
		ar.setName(name);
		addAdministrativeDomainInfo(ar, domAr);
		return ar;
	}


	private Identifier transformDev(Element domDev) throws UnmarshalException {
		Device dev = mIdentifierFactory.createDev();
		Element child = null;
		if (domDev.getChildNodes().getLength() != 1) 
			throw new UnmarshalException("Bad " + IfmapStrings.DEVICE_EL_NAME +
					" element?");
		
		if (!(domDev.getFirstChild().getNodeType() == Node.ELEMENT_NODE))
			throw new UnmarshalException("Element " + IfmapStrings.DEVICE_EL_NAME +
					"has no child element");
		
		child = (Element)domDev.getFirstChild();
		
		if (!elementMatches(child, IfmapStrings.DEVICE_NAME_EL_NAME))
			throw new UnmarshalException("Unknown child element in " 
					+ IfmapStrings.DEVICE_EL_NAME + " element");
		
		if (child.getFirstChild().getNodeType() != Node.TEXT_NODE)
			throw new UnmarshalException("No text node found");
		
		String name = child.getFirstChild().getTextContent();
		
		if (name == null)
			throw new UnmarshalException("No text content found");
		
		dev.setName(name);
		return dev;
	}

	private Identifier transformIdentity(Element domId) throws UnmarshalException {
		Identity id = mIdentifierFactory.createIdentity();
		IdentityType typeEnum = null;
		String name = domId.getAttribute(IfmapStrings.IDENTITY_ATTR_NAME);
		String type = domId.getAttribute(IfmapStrings.IDENTITY_ATTR_TYPE);
		String otherTypeDef = 
			domId.getAttribute(IfmapStrings.IDENTITY_ATTR_OTHER_TYPE_DEF);
		
		if (name == null)
			throw new UnmarshalException("name not set fo " +
					IfmapStrings.IDENTITY_EL_NAME);
		
		for (IdentityType t : IdentityType.values()) {
			if (t.toString().equals(type)) {
				typeEnum = t;
				break;
			}
		}
		
		if (typeEnum == null)
			throw new UnmarshalException("No type for " + 
					IfmapStrings.IDENTITY_EL_NAME + " given");
		
		if (typeEnum == IdentityType.other && otherTypeDef == null)
			throw new UnmarshalException(IfmapStrings.IDENTITY_EL_NAME +
					" with type other, but not other type definition");
		
		id.setName(name);
		id.setType(typeEnum);
		id.setOtherTypeDefinition(otherTypeDef);
		
		addAdministrativeDomainInfo(id, domId);
		return id;
	}

	private Identifier transformIp(Element domIp) throws UnmarshalException {
		IpAddress ip = null;
		String value = domIp.getAttribute(IfmapStrings.IP_ADDRESS_ATTR_VALUE);
		String type = domIp.getAttribute(IfmapStrings.IP_ADDRESS_ATTR_TYPE);
		IpAddressType typeEnum = IpAddressType.IPv4;
		
		if (value == null)
			throw new UnmarshalException("No value for " +
					IfmapStrings.IP_ADDRESS_EL_NAME + " found");
		
		if (type != null) {
			try {
			typeEnum = IpAddressType.valueOf(type);
			} catch (IllegalArgumentException e) {
				throw new UnmarshalException("Invalid type for " 
						+ IfmapStrings.IP_ADDRESS_EL_NAME + " found");
			}
		}
		
		ip = mIdentifierFactory.createIp(typeEnum, value, null);
		addAdministrativeDomainInfo(ip, domIp);
		
		return ip;
	}

	private Identifier transformMac(Element domMac) throws UnmarshalException {
		MacAddress mac = mIdentifierFactory.createMac();
		String value = domMac.getAttribute(IfmapStrings.MAC_ADDRESS_ATTR_VALUE);
		if (value == null)
			throw new UnmarshalException("No value for " +
					IfmapStrings.MAC_ADDRESS_EL_NAME + " found");
		
		mac.setValue(value);
		addAdministrativeDomainInfo(mac, domMac);
		return mac;
	}

	private List<Document> transformMetadata(Element mdElement) throws UnmarshalException {
		NodeList children = mdElement.getChildNodes();
		List<Document> ret = new LinkedList<Document>();
		
		
		for (int i = 0; i < children.getLength(); i++) {
			Document mdDoc = null;
			if (!(children.item(i).getNodeType() == Node.ELEMENT_NODE))
				continue;
			
			mdDoc = mDocumentBuilder.newDocument();
			Element el = (Element) mdDoc.importNode(children.item(i), true);
			mdDoc.appendChild(el);
			ret.add(mdDoc);
		}
		
		return ret;
	}

	private void addAdministrativeDomainInfo(IdentifierWithAd id, Element child) {
		String admDom = child.getAttribute(IfmapStrings.IDENTIFIER_ATTR_ADMIN_DOMAIN);
		if (admDom != null)
			id.setAdministrativeDomain(admDom);
	}
	
	

	private PollResult parsePollResultElement(Element e) throws EndSessionException,
			UnmarshalException, IfmapErrorResult {
		
		Collection<SearchResult> sres = new LinkedList<SearchResult>();
		Collection<SearchResult> ures = new LinkedList<SearchResult>();
		Collection<SearchResult> dres = new LinkedList<SearchResult>();
		Collection<SearchResult> nres = new LinkedList<SearchResult>();
		Collection<IfmapErrorResult> eres = new LinkedList<IfmapErrorResult>();
		
		if (elementMatches(e, IfmapStrings.POLL_RES_EL_NAME)) {
			NodeList children = e.getChildNodes();
			
			for (int i = 0; i < children.getLength(); i++) {
				SearchResult sr = null;
				boolean isSearchResult = true;
				Collection<SearchResult> addHere = null;
				
				if (!(children.item(i).getNodeType() == Node.ELEMENT_NODE))
					continue;
				
				Element child = (Element)children.item(i);
				
				
				if (elementMatches(child, IfmapStrings.POLL_SEARCH_RES_EL_NAME)) {
					addHere = sres;
				} else if (elementMatches(child, IfmapStrings.POLL_UPDATE_RES_EL_NAME)) {
					addHere = ures;
				} else if (elementMatches(child, IfmapStrings.POLL_DELETE_RES_EL_NAME)) {
					addHere = dres;
				} else if (elementMatches(child, IfmapStrings.POLL_NOTIFY_RES_EL_NAME)) {
					addHere = nres;
				} else if (elementMatches(child, IfmapStrings.ERROR_RES_EL_NAME)) {
					isSearchResult = false;
				} else {
					throw new UnmarshalException("Unknown element in " +
							IfmapStrings.POLL_RES_EL_NAME);
				}
				
				if (isSearchResult) {
					sr = transformSearchResult(child);
					if (sr.getName() == null)
						throw new UnmarshalException("No name attribute for result");
					
					addHere.add(sr);
				} else {
					IfmapErrorResult err = parseErrorResult(child);
					eres.add(err);
				}
			}
			
		} else if (elementMatches(e, IfmapStrings.END_SESSION_RES_EL_NAME)) {
			throw new EndSessionException();
		} else {
			checkForError(e);
			// if we get here, we didn't find a pollResult, no endSessionResult
			// and no errorResult so something is weird ;)
			throw new UnmarshalException("Bad response poll response");
		}
		return mResultFactory.createPollRes(sres, ures, dres, nres, eres);
	}

	@Override
	public void checkEndSessionResult(InputStream is) throws IfmapErrorResult,
			UnmarshalException {
		checkResult(is, IfmapStrings.END_SESSION_RES_EL_NAME);
	}

	@Override
	public void checkRenewSessionResult(InputStream is)
			throws IfmapErrorResult, UnmarshalException {
		checkResult(is, IfmapStrings.RENEW_SESSION_RES_EL_NAME);
	}

	@Override
	public void checkPurgePublisherReceived(InputStream is)
			throws IfmapErrorResult, UnmarshalException {
		checkResult(is, IfmapStrings.PURGE_PUBLISHER_RES_EL_NAME);
	}

	@Override
	public void checkPublishReceived(InputStream is) throws IfmapErrorResult,
			UnmarshalException {
		checkResult(is, IfmapStrings.PUBLISH_RES_EL_NAME);
	}

	@Override
	public void checkSubscribeReceived(InputStream is) throws IfmapErrorResult,
			UnmarshalException {
		checkResult(is, IfmapStrings.SUBSCRIBE_RES_EL_NAME);
	}


	private NewSessionResult parseNewSessionResultElement(Element elem) 
			throws UnmarshalException {
		String sId, pId, mprs;
		Integer mprsVal = null;
		if (!elementMatches(elem, IfmapStrings.NEW_SESSION_RES_EL_NAME))
			throw new UnmarshalException("No newSession element found");
			
		sId = elem.getAttribute( IfmapStrings.NEW_SESSION_RES_ATTR_SID);
		pId = elem.getAttribute(IfmapStrings.NEW_SESSION_RES_ATTR_PID);
		mprs = elem.getAttribute(IfmapStrings.NEW_SESSION_RES_ATTR_MPRS);
			
		if (sId == null || sId.length() == 0)
			throw new UnmarshalException("No " + 
					IfmapStrings.NEW_SESSION_RES_ATTR_SID + " found");
		if (pId == null || sId.length() == 0)
			throw new UnmarshalException("No " + 
					IfmapStrings.NEW_SESSION_RES_ATTR_PID + " found");
		
		if (mprs != null && mprs.length() > 0) {
			try {
				mprsVal = Integer.parseInt(mprs);
			} catch (NumberFormatException e) {
				throw new UnmarshalException("error parsing max-poll-result-size " 
						+ "[" + mprs + "]");
			}
		}
			
		return mResultFactory.createNewSessionRes(sId, pId, mprsVal);
	}


	private Element findResponseElement(Document doc) throws UnmarshalException {
		Element env, body, response;
		env = findElementInChildren(doc, IfmapStrings.SOAP_ENV_EL_NAME,
				IfmapStrings.SOAP_ENV_NS_URI);
		
		if (env == null)
			throw new UnmarshalException("Could not find SOAP Envelope");
		
		body = findElementInChildren(env, IfmapStrings.SOAP_BODY_EL_NAME,
				IfmapStrings.SOAP_ENV_NS_URI);

		if (body == null)
			throw new UnmarshalException("Could not find SOAP Body");
		
		response = findElementInChildren(body, IfmapStrings.RESPONSE_EL_NAME,
				IfmapStrings.BASE_NS_URI);

		if (response == null)
			throw new UnmarshalException("Could not find IF-MAP Response");
		
		return response;
	}
	

	private Document parseDocument(InputStream is) throws UnmarshalException {
		try {
			return mDocumentBuilder.parse(is);
		} catch (Exception e) { // catchall, not nice but we don't do anything special
			IfmapJLog.error("Parsing Exception occurred [" + e.getMessage() +"]");
			throw new UnmarshalException(e.getMessage());
		}
	}

	
	private void checkForError(Element element) throws UnmarshalException, IfmapErrorResult {
		if (elementMatches(element, IfmapStrings.ERROR_RES_EL_NAME))
			throw parseErrorResult(element);
	}


	private IfmapErrorResult parseErrorResult(Element element) throws UnmarshalException {
		String errCodeStr, errStr;
		Element errStrElem = null;
		IfmapErrorCode errCode = null;
		errCodeStr = element.getAttribute( IfmapStrings.ERROR_RES_ATTR_ERRCODE);
			
		errStrElem = findElementInChildren(element, IfmapStrings.ERROR_RES_ERRSTR_EL_NAME,
				IfmapStrings.NO_URI);
			
		errStr = (errStrElem == null) ? 
				"IfmapJ: errorString not given" : errStrElem.getTextContent();
		
		try {
			errCode = IfmapErrorCode.valueOf(errCodeStr);
		} catch (IllegalArgumentException e) {
			throw new UnmarshalException("Invalid errorCode received: " +
					errCodeStr);
		}
		return new IfmapErrorResult(errCode, errStr);
	}

	private boolean elementMatches(Element el, String elname) {
		return elementMatches(el, elname, IfmapStrings.NO_URI);
	}
	
	private boolean elementMatches(Element e, String name, String uri) {
		// i'm sorry
		return (name.equals(e.getLocalName())
				&& ((uri == null && e.getNamespaceURI() == null)
					|| (uri != null && uri.equals(e.getNamespaceURI()))));
	}

	
	private Element findElementInChildren(Node node, String name, String uri) {
		NodeList children = node.getChildNodes();
		Element ret = null;
		for (int i = 0; i < children.getLength(); i++) {
			Node n = children.item(i);
			
			// our interest is in element only
			if (!(n.getNodeType() == Node.ELEMENT_NODE))
				continue;
				
			if (elementMatches((Element)n, name, uri)) {
				ret = (Element)n;
				break;
			}
		}
		return ret;
	}

	
	private Element getResponseContentErrorCheck(InputStream is) throws UnmarshalException, IfmapErrorResult {
		Node searchNode = null;
		Element responseContent = null;
		Document doc = parseDocument(is);
		Element response = findResponseElement(doc);
		NodeList children = response.getChildNodes();
		if (children.getLength() == 0)
			throw new UnmarshalException("No element in response element found");
		
	
		// find a element element and return it...
		
		for (int i = 0; i < children.getLength(); i++) {
			searchNode = children.item(i);
			if (searchNode.getNodeType() == Node.ELEMENT_NODE) {
				if (responseContent != null)
					throw new UnmarshalException("multiple elements in response " +
							"element found");
				else
					responseContent = (Element)searchNode;
			}
			
		}
		
		if (responseContent == null)
			throw new UnmarshalException("No element found in response element");
		
		checkForError(responseContent);
		return responseContent;
	}

	private void checkResult(InputStream is, String expected)
			throws UnmarshalException, IfmapErrorResult {
		Element el = getResponseContentErrorCheck(is);
		
		if (!elementMatches(el, expected))
			throw new UnmarshalException("No " + expected + " element found");
		
		if (el.getChildNodes().getLength() != 0)
			throw new UnmarshalException("Unexpected children for " + expected +
					" element");
	}
}
