package de.fhhannover.inform.trust.ifmapj.channel;

import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.zip.GZIPInputStream;

import javax.net.ssl.SSLSocket;
import javax.net.ssl.SSLSocketFactory;

import org.apache.http.Header;
import org.apache.http.HttpEntity;
import org.apache.http.HttpException;
import org.apache.http.HttpResponse;
import org.apache.http.StatusLine;
import org.apache.http.entity.InputStreamEntity;
import org.apache.http.impl.DefaultHttpClientConnection;
import org.apache.http.message.BasicHttpEntityEnclosingRequest;
import org.apache.http.params.BasicHttpParams;

import util.Base64;
import de.fhhannover.inform.trust.ifmapj.exception.CommunicationException;
import de.fhhannover.inform.trust.ifmapj.exception.InitializationException;
import de.fhhannover.inform.trust.ifmapj.log.IfmapJLog;

/**
 * Implementation of the {@link CommunicationHandler} interface using
 * Apache's httpcomponent-core package only.
 * 
 * @author aw
 *
 */
class ApacheCoreCommunicationHandler implements CommunicationHandler {
	
	
	private final BasicHttpParams mBasicHttpParams;
	private final String mUser;
	private final String mPass;
	private final URL mUrl;
	private final String mPath;
	private final int mPort;
	private final SSLSocketFactory mSocketFactory;
	
	private boolean mUseGzip;
	private final boolean mBasicAuth;

	private SSLSocket mSocket;
	private DefaultHttpClientConnection mHttpConnection;
	
	ApacheCoreCommunicationHandler(String url, String user, String pass,
			SSLSocketFactory sslSocketFactory) throws InitializationException {
		
		if (url == null)
			throw new NullPointerException("url is null");
		
		if (sslSocketFactory == null)
			throw new NullPointerException("sslSocketFactory is null");
		
		try {
			mUrl = new URL(url);
		} catch (MalformedURLException e) {
			throw new InitializationException("Malformed URL [" 
					+ e.getMessage() + "]");
		}

		// fix up the path in case non is given
		mPath = (mUrl.getPath().length() == 0) ? "/" : mUrl.getPath();
	
		mPort = (mUrl.getPort() == -1) ? mUrl.getDefaultPort() : mUrl.getPort();
		
		if (mPort < 0)
			throw new InitializationException("Could not determine port to use");
		
		mBasicHttpParams = new BasicHttpParams();
		mBasicAuth = (user != null && pass != null);
		mUser = user;
		mPass = pass;
		mSocketFactory = sslSocketFactory;
	}

	@Override
	public InputStream doRequest(InputStream is) throws CommunicationException {
		try {
			boolean respIsGzip = false;
			InputStreamEntity ise = null;
			HttpResponse response = null;
			InputStream ret = null;
			BasicHttpEntityEnclosingRequest httpPost = null;
			Header hdr = null;
			HttpEntity respEntity = null;
			StatusLine status = null;
		
			// make sure this stuff is up and running
			if (mSocket == null) {
				// clean up first, if mSocket and mHttpConnection are null, this is
				// a no-op
				closeTcpConnection();
				mSocket = (SSLSocket) mSocketFactory.createSocket(mUrl.getHost(), mPort);
				mSocket.setNeedClientAuth(true);
			}
			
			if (mHttpConnection == null) {
				mHttpConnection = new DefaultHttpClientConnection();
				mHttpConnection.bind(mSocket, mBasicHttpParams);
			}
	
			httpPost = new BasicHttpEntityEnclosingRequest("POST", mPath);
		
			// add some headers
			httpPost.addHeader("Content-type", "application/soap+xml");
			httpPost.addHeader("Accept", "application/soap+xml");
	
			// prepare gzip compressed content and headers if we use gzip
			if (usesGzip()) {
				httpPost.addHeader("Accept-Encoding", "gzip");
				httpPost.addHeader("Content-Encoding", "gzip");
			}
	
			// authentication information when basic auth is used
			if (isBasicAuth())
				httpPost.addHeader("Authorization", "Basic " +
						Base64.encodeToString(((mUser + ":" + mPass).getBytes()), false));
	
	
	
			// add the entity to the post request
			httpPost.addHeader("Content-Length", "" + is.available());
			ise = new InputStreamEntity(is, is.available());
			ise.setChunked(false);
			httpPost.setEntity(ise);
		
			// do the actual request
			try {
				mHttpConnection.sendRequestHeader(httpPost);
				mHttpConnection.sendRequestEntity(httpPost);
				response = mHttpConnection.receiveResponseHeader();
				mHttpConnection.receiveResponseEntity(response);
			} catch (HttpException e) {
				IfmapJLog.warn("HTTP request failed [" + e.getMessage() + "]");
				throw new IOException(e);
			}
	
			// check if we got a 200 status back, otherwise go crazy
			status = response.getStatusLine();
			if (status.getStatusCode() != 200) {
				IfmapJLog.warn("HTTP Status Code: " 
						+ status.getStatusCode() + " "
						+ status.getReasonPhrase());
				throw new CommunicationException("HTTP Status Code: "
						+ status.getStatusCode() + " " 
						+ status.getReasonPhrase());
			}
			
			
			// check whether the response uses gzip
			hdr = response.getFirstHeader("Content-Encoding");
			respIsGzip = (hdr == null) ? false : hdr.getValue().contains("gzip");
			respEntity = response.getEntity();
			
			if (respEntity != null)
				ret = respEntity.getContent();
			else 
				throw new CommunicationException("no content in response");
			
			if (respIsGzip) {
				ret = new GZIPInputStream(ret);
			}
			
			return ret;
		} catch (IOException e) {
			throw new CommunicationException(e);
		}
	}
	
	@Override
	public void closeTcpConnection() throws CommunicationException {
		Exception tmpException = null;
		
		try {
			if (mSocket != null) mSocket.close();
		} catch (IOException e) {
			tmpException = e;
		} finally {
			mSocket = null;
		}
		
		try {
			if (mHttpConnection != null) mHttpConnection.close();
		} catch (IOException e) {
			if (tmpException == null) tmpException = e;
		} finally {
			mHttpConnection = null;
		}
		
		if (tmpException != null)
			throw new CommunicationException(tmpException);
	}


	public boolean isBasicAuth() {
		return mBasicAuth;
	}

	@Override
	public void setGzip(boolean gzip) {
		mUseGzip = gzip;
	}


	@Override
	public boolean usesGzip() {
		return mUseGzip;
	}
}
