package de.fhhannover.inform.trust.ifmapj.metadata;

import org.w3c.dom.Document;

/**
 * Inteface to create some of the standard metadata specified by IF-MAP 2.0.
 * 
 * TODO: This is not complete. The more complex metadata types are missing.
 *        For example, device-characteristic and event are not there yet.
 * 
 * @author aw
 *
 */
public interface StandardIfmapMetadataFactory {
	
	/**
	 * Create a {@link Document} representing ip-mac metadata.
	 * 
	 * <pre>
	 *  ip-mac is link metadata that associates an
     *  ip-address identifier with a mac-address identifier
     *  and which includes optional DHCP lease information
     * <\pre>
	 * 
	 * @param startTime the start-time value or null
	 * @param endTime  the end-time value or null
	 * @param dhcpServer the dhcp-server value or null
	 * @return
	 */
	public Document createIpMac(String startTime, String endTime, String dhcpServer);

	/**
	 * Create a {@link Document} representing ip-mac metadata.
	 * 
	 * <pre>
	 *  ip-mac is link metadata that associates an
     *  ip-address identifier with a mac-address identifier
     *  and which includes optional DHCP lease information
     * <\pre>
	 * 
	 * @return
	 */
	public Document createIpMac();
	

	/**
	 * Create a {@link Document} representing access-request-mac metadata.
	 * 
	 * <pre>
	 *  access-request-mac is link metadata that
     *  associates an access-request identifier with
     *  a mac-address identifier 
     * </pre>
	 * @return
	 */
	public Document createArMac();
	
	/**
	 * Create a {@link Document} representing access-request-device metadata.
	 * 
	 * <pre>
	 *  access-request-device is link metadata that
     *  associates an access-request identifier with
     *  a device identifier
     * </pre>
	 * 
	 * @return
	 */
	public Document createArDev();
	
	/**
	 * Create a {@link Document} representing access-request-ip metadata.
	 * 
	 * <pre>
	 *  access-request-ip is link metadata that
     *  associates an access-request identifier with
     *  an ip-address identifier
     * </pre>
	 * 
	 * @return
	 */
	public Document createArIp();
	
	/**
	 * Create a {@link Document} representing authenticated-as metadata.
	 * 
	 * <pre>
	 *  authenticated-as is link metadata that
     *  associates an access-request identifier with
     *  an identity identifier
     * </pre>
	 * 
	 * @return
	 */
	public Document createAuthAs();

	/**
	 * Create a {@link Document} representing authenticated-by metadata.
	 * 
	 * <pre>
	 *  authenticated-by is link metadata that
     *  associates an access-request identifier with
     *  the device identifier of the PDP that
     *  authenticated the access-request 
	 * </pre>
	 * 
	 * @return
	 */
	public Document createAuthBy(); 
	
	
	/**
	 * Create a {@link Document representing device-ip metadata.
	 * 
	 * <pre>
	 *  device-ip is link metadata that associates a device
     *  identifier of a PDP with an IP address which it has
     *  authenticated 
	 * </pre>
	 * 
	 * @return
	 */
	public Document createDevIp();

	/**
	 * Create a {@link Document} representing discovered-by metadata.
	 * 
	 * <pre>
	 *  discovered-by is link metadata that associates
     *  an ip-address or mac-address identifier of an endpoint
     *  with the device identifier of a MAP Client that has
     *  noticed the endpoint on the network 
	 * </pre>
	 * @return
	 */
	public Document createDiscoveredBy();
	
	
	/**
	 * Create a {@link Document} representing role metadata.
	 * 
	 * <pre>
	 *  role is link metadata that associates an
     *  access-request identifier with an identity
     *  identifier and which names collections of
     *  privileges associated with the end-user
	 * </pre>
	 * 
	 * @param name the name of the role
	 * @param administrativeDomain the administrative-domain or null
	 * 
	 * @return
	 */
	public Document createRole(String name, String administrativeDomain);

	/**
	 * Create a {@link Document} representing role metadata.
	 * 
	 * <pre>
	 *  role is link metadata that associates an
     *  access-request identifier with an identity
     *  identifier and which names collections of
     *  privileges associated with the end-user
	 * </pre>
	 * 
	 * @param name
	 * @return
	 */
	public Document createRole(String name);


	/**
	 * Create a {@link Document} representing device-attribute metadata.
	 * 
	 * <pre>
	 *  device-attribute is link metadata that associates
     *  an access-request identifier with a device identifier
     *  and which includes information about the device such
     *  as its health 
	 * </pre>
	 * 
	 * @param name the name of the attribute
	 * @return
	 */
	public Document createDevAttr(String name);

}
